--------------------------------------------------------------------------------
-- House of Winters - Research and Mount Timers
--------------------------------------------------------------------------------
-- Monitor your crafting timers and mount feeding schedule.
--------------------------------------------------------------------------------
-- Exposes the following to the global scope:
--     HoW_RaMT             (addon namespace)
--     HoW_RaMT_CT_*        (addon controls)
--     HoW_RaMT_vars        (saved variables)
--
-- Provides the following slash commands:
--     /timers
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup addon global variable.
--------------------------------------------------------------------------------

HoW_RaMT = {
    name         = "HoW_RaMT",
    displayName  = "Research and Mount Timers",
    abbrName     = "R&M Timers",
    prefix       = "HoW_RaMT_CT_",
    author       = "Kentarii",
    version      = "1.2.0",
    link         = "http://www.esoui.com/",
    command      = "/timers",
    types        = {
        "Blacksmithing",
        "Clothing",
        "Woodworking",
        "Mounts"
    },
}

--------------------------------------------------------------------------------
-- Libraries, defaults and constants.
--------------------------------------------------------------------------------

local lam = LibStub:GetLibrary("LibAddonMenu-1.0")
local howui = LibStub:GetLibrary("LibHoWUI-1.0")

--
-- AddOn defaults.
--
local defaults = {
    language = "en",
    names = {},
    chars = {},
    colors = {
        header = "EECA2A",
        section = "FF9933",
        text = "FFE3A1",
        value = "F0F0F0",
        info = "008000",
        alert = "C00000",
    },
}

--
-- Character defaults.
--
local charDefaults = {
    visible = true,
    faded = false,
    alpha = 65,
    anchor = {
        point = TOPLEFT,
        relativePoint = TOPLEFT,
        offsetX = 10,
        offsetY = 60,
    },
    dimension = {
        width = 250,
        height = 500,
    },
    unit = {
        name = nil,
        level = nil,
        race = nil,
        class = nil,
    },
    types = {
        Blacksmithing = {
            maxSlots = 0,
            timers = {},
            visible = true,
        },
        Clothing = {
            maxSlots = 0,
            timers = {},
            visible = true,
        },
        Woodworking = {
            maxSlots = 0,
            timers = {},
            visible = true,
        },
        Mounts = {
            maxSlots = 0,
            timers = {},
            visible = true,
        }
    },
}

--
-- Used for building data structure, should always be in english.
--
local KEY_LANG = "en"

--
-- Dictionary for localization.
--
local MOUNTS_TYPE = {
    ["en"] = "Mounts"
}

--
-- Map crafting type constants to text.
--
local CRAFTING_TYPES = {
    ["en"] = {
        [CRAFTING_TYPE_ALCHEMY]       = "Alchemy",
        [CRAFTING_TYPE_BLACKSMITHING] = "Blacksmithing",
        [CRAFTING_TYPE_CLOTHIER]      = "Clothing",
        [CRAFTING_TYPE_ENCHANTING]    = "Enchanting",
        [CRAFTING_TYPE_INVALID]       = "Invalid",
        [CRAFTING_TYPE_PROVISIONING]  = "Provisioning",
        [CRAFTING_TYPE_WOODWORKING]   = "Woodworking"
    }
}

--
-- Map item trait type constants to text.
--
local ITEM_TRAIT_TYPES = {
    ["en"] = {
        [ITEM_TRAIT_TYPE_ARMOR_DIVINES]      = "Divines",
        [ITEM_TRAIT_TYPE_ARMOR_EXPLORATION]  = "Exploration",
        [ITEM_TRAIT_TYPE_ARMOR_IMPENETRABLE] = "Impenetrable",
        [ITEM_TRAIT_TYPE_ARMOR_INFUSED]      = "Infused",
        [ITEM_TRAIT_TYPE_ARMOR_INTRICATE]    = "Intricate",
        [ITEM_TRAIT_TYPE_ARMOR_ORNATE]       = "Ornate",
        [ITEM_TRAIT_TYPE_ARMOR_REINFORCED]   = "Reinforced",
        [ITEM_TRAIT_TYPE_ARMOR_STURDY]       = "Sturdy",
        [ITEM_TRAIT_TYPE_ARMOR_TRAINING]     = "Training",
        [ITEM_TRAIT_TYPE_ARMOR_WELL_FITTED]  = "Well-fitted",
        [ITEM_TRAIT_TYPE_JEWELRY_ARCANE]     = "Arcane",
        [ITEM_TRAIT_TYPE_JEWELRY_HEALTHY]    = "Healthy",
        [ITEM_TRAIT_TYPE_JEWELRY_ORNATE]     = "Ornate",
        [ITEM_TRAIT_TYPE_JEWELRY_ROBUST]     = "Robust",
        [ITEM_TRAIT_TYPE_NONE]               = "None",
        [ITEM_TRAIT_TYPE_WEAPON_CHARGED]     = "Charged",
        [ITEM_TRAIT_TYPE_WEAPON_DEFENDING]   = "Defending",
        [ITEM_TRAIT_TYPE_WEAPON_INFUSED]     = "Infused",
        [ITEM_TRAIT_TYPE_WEAPON_INTRICATE]   = "Intricate",
        [ITEM_TRAIT_TYPE_WEAPON_ORNATE]      = "Ornate",
        [ITEM_TRAIT_TYPE_WEAPON_POWERED]     = "Powered",
        [ITEM_TRAIT_TYPE_WEAPON_PRECISE]     = "Precise",
        [ITEM_TRAIT_TYPE_WEAPON_SHARPENED]   = "Sharpened",
        [ITEM_TRAIT_TYPE_WEAPON_TRAINING]    = "Training",
        [ITEM_TRAIT_TYPE_WEAPON_WEIGHTED]    = "Weighted",
    }
}

--
-- Supported fonts
--
local FONTS = {
   -- "ZoFontBookLetter",
   -- "ZoFontBookNote",
   -- "ZoFontBookPaper",
   -- "ZoFontBookRubbing",
   -- "ZoFontBookScroll",
   -- "ZoFontBookSkin",
   -- "ZoFontBookTablet",
   "ZoFontChat",
   "ZoFontGame",
   "ZoFontGameBold",
   "ZoFontGameLarge",
   "ZoFontGameLargeBold",
   "ZoFontGameLargeBoldShadow",
   "ZoFontGameMedium",
   "ZoFontGameOutline",
   "ZoFontGameShadow",
   "ZoFontGameSmall",
   -- "ZoFontWinT1",
   -- "ZoFontWinT2",
}

--------------------------------------------------------------------------------
-- Addon init and event/command handling functions.
--------------------------------------------------------------------------------

--- Event handler for EVENT_ADD_ON_LOADED.
-- @return void
function HoW_RaMT.AddOnLoaded(eventCode, addOnName)
    if addOnName ~= HoW_RaMT.name then return end

    -- Register the slash command handler.
    SLASH_COMMANDS[HoW_RaMT.command] = HoW_RaMT.CommandHandler

    -- Initialize addon.
    HoW_RaMT:Init()

    -- Add addon settings page.
    HoW_RaMT.Settings()
end

--- Initialize addon.
-- @return void
function HoW_RaMT:Init()
    -- Load saved vars.
    HoW_RaMT.gvars = ZO_SavedVars:NewAccountWide("HoW_RaMT_vars", 1, HoW_RaMT.name, defaults)
    HoW_RaMT.unitName = GetUnitName("player")

    if not HoW_RaMT.gvars.chars[HoW_RaMT.unitName] then
        -- Set character defaults
        table.insert(HoW_RaMT.gvars.names, HoW_RaMT.unitName)
        HoW_RaMT.gvars.chars[HoW_RaMT.unitName] = charDefaults
        HoW_RaMT.gvars.chars[HoW_RaMT.unitName].unit.name  = HoW_RaMT.unitName
        HoW_RaMT.gvars.chars[HoW_RaMT.unitName].unit.race  = GetUnitRace("player")
        HoW_RaMT.gvars.chars[HoW_RaMT.unitName].unit.class = GetUnitClass("player")
    end
    HoW_RaMT.vars = HoW_RaMT.gvars.chars[HoW_RaMT.unitName]

    -- Update data which may have changed
    HoW_RaMT.vars.unit.level = GetUnitLevel("player")
    HoW_RaMT.vars.unit.rank = GetUnitVeteranRank("player")
    HoW_RaMT.vars.types.Blacksmithing.maxSlots = GetMaxSimultaneousSmithingResearch(CRAFTING_TYPE_BLACKSMITHING)
    HoW_RaMT.vars.types.Clothing.maxSlots      = GetMaxSimultaneousSmithingResearch(CRAFTING_TYPE_CLOTHIER)
    HoW_RaMT.vars.types.Woodworking.maxSlots   = GetMaxSimultaneousSmithingResearch(CRAFTING_TYPE_WOODWORKING)
    HoW_RaMT.vars.types.Mounts.maxSlots        = GetMaxStableSlots()

    -- Update control from vars.
    HoW_RaMT:AddCharacterButtons()
    HoW_RaMT.SetAlpha(HoW_RaMT.vars.alpha)
    HoW_RaMT:SetAnchor(HoW_RaMT.vars.anchor)
    HoW_RaMT:SetDimensions(HoW_RaMT.vars.dimension)
    HoW_RaMT:SetFaded(HoW_RaMT.vars.faded)

    -- Load first time data and save
    HoW_RaMT.Update()

    if HoW_RaMT.vars.visible then
        HoW_RaMT:OpenWindow()
    end

    -- Bind event handlers.
    EVENT_MANAGER:RegisterForEvent(HoW_RaMT.name, EVENT_SMITHING_TRAIT_RESEARCH_STARTED,   HoW_RaMT.OnSmithingTraitResearchStarted)
    EVENT_MANAGER:RegisterForEvent(HoW_RaMT.name, EVENT_SMITHING_TRAIT_RESEARCH_COMPLETED, HoW_RaMT.OnSmithingTraitResearchCompleted)
    EVENT_MANAGER:RegisterForEvent(HoW_RaMT.name, EVENT_STABLE_INTERACT_END,               HoW_RaMT.OnStableInteractEnd)

    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnShow", function()
        HoW_RaMT:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_GameMenu_InGame, "OnHide", function()
        HoW_RaMT:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnShow", function()
        HoW_RaMT:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_InteractWindow, "OnHide", function()
        HoW_RaMT:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnShow", function()
        HoW_RaMT:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_KeybindStripControl, "OnHide", function()
        HoW_RaMT:ShowIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnShow", function()
        HoW_RaMT:HideIfVisible()
    end)
    ZO_PreHookHandler(ZO_MainMenuCategoryBar, "OnHide", function()
        HoW_RaMT:ShowIfVisible()
    end)
end

--- Event handler for EVENT_SMITHING_TRAIT_RESEARCH_STARTED.
-- @param eventCode
-- @param craftingSkillType
-- @param researchLineIndex
-- @param traitIndex
-- @return void
function HoW_RaMT.OnSmithingTraitResearchStarted(eventCode, craftingSkillType, researchLineIndex, traitIndex)
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][craftingSkillType]].timers = HoW_RaMT.GetCraftingTimers(craftingSkillType)
end

--- Event handler for EVENT_SMITHING_TRAIT_RESEARCH_COMPLETED.
-- @param craftingSkillType
-- @param researchLineIndex
-- @param traitIndex
-- @return void
function HoW_RaMT.OnSmithingTraitResearchCompleted(eventCode, craftingSkillType, researchLineIndex, traitIndex)
    local name, icon, numTraits, timeRequiredForNextResearchSecs = GetSmithingResearchLineInfo(craftingSkillType, researchLineIndex)
    local traitType, traitDescription, known = GetSmithingResearchLineTraitInfo(craftingSkillType, researchLineIndex, traitIndex)
    d(string.format("[%s] Finished researching trait %s for %s.", CRAFTING_TYPES[HoW_RaMT.gvars.language][craftingSkillType], ITEM_TRAIT_TYPES[HoW_RaMT.gvars.language][traitType], name))
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][craftingSkillType]].timers = HoW_RaMT.GetCraftingTimers(craftingSkillType)
end

--- Event handler for EVENT_STABLE_INTERACT_END.
-- @return void
function HoW_RaMT.OnStableInteractEnd()
    HoW_RaMT.UpdateMounts()
end

--- Event handler for OnMoveStop to reposition window per character.
-- @return void
function HoW_RaMT:OnMoveStop()
    local isValidAnchor, point, relativeTo, relativePoint, offsetX, offsetY = HoW_RaMT_CT:GetAnchor()
    local vars = HoW_RaMT.gvars.chars[HoW_RaMT.unitName]
    if (isValidAnchor) then
        vars.anchor.point = point
        vars.anchor.relativePoint = relativePoint
        vars.anchor.offsetX = offsetX
        vars.anchor.offsetY = offsetY
    end
end

--- Event handler for OnResizeStop to resize window per character.
-- @return void
function HoW_RaMT:OnResizeStop()
    local width, height = HoW_RaMT_CT:GetDimensions()
    local vars = HoW_RaMT.gvars.chars[HoW_RaMT.unitName]
    if width and height then
        vars.dimension.width = width
        vars.dimension.height = height
    end
    -- Resize might cause anchor change as well.
    HoW_RaMT:OnMoveStop()
end

--- Addon command handler.
-- @param text the command line argument (original case)
-- @return void
function HoW_RaMT.CommandHandler(text)
    local cmd = string.lower(text)

    if cmd == "show" then
        HoW_RaMT.Update()
        HoW_RaMT.OpenWindow()
    elseif cmd == "hide" then
        HoW_RaMT:CloseWindow()
    else
        d("----------------------------------------")
        d("|cEECA2A"..HoW_RaMT.displayName.."|r |c6699FF"..HoW_RaMT.version.."|r by |cFF9933"..HoW_RaMT.author.."|r")
        d("----------------------------------------")
        d("Commands:")
        d(HoW_RaMT.command.." show")
        d(HoW_RaMT.command.." hide")
        d("----------------------------------------")
    end
end

--------------------------------------------------------------------------------
-- Setup addon control panel page.
--------------------------------------------------------------------------------

--- Addon settings.
-- @return void
function HoW_RaMT.Settings()
    local controlPanel = HoW_Settings

    -- Create HoW settings page.
    if not controlPanel then
        controlPanel = lam:CreateControlPanel("HoW_Settings", "|cEECA2AHoW Addons|r")
    end

    -- Add module settings.
    lam:AddHeader(controlPanel, HoW_RaMT.prefix.."SettingsHeader", HoW_RaMT.displayName)
    lam:AddCheckbox(controlPanel, HoW_RaMT.prefix.."SettingsVisible", "Visible", "Toggle Texture Viewer visibility", HoW_RaMT.IsVisible, HoW_RaMT.SetVisible)
    --lam:AddDropdown(controlPanel, HoW_RaMT.prefix.."SettingsFont", "Font", "Change font family", FONTS, HoW_RaMT.GetFont, HoW_RaMT.SetFont)
    lam:AddSlider(controlPanel, HoW_RaMT.prefix.."SettingsAlpha", "Alpha", "Change background transparency", 0, 100, 1, HoW_RaMT.GetAlpha, HoW_RaMT.SetAlpha)
    lam:AddDescription(controlPanel, HoW_RaMT.prefix.."AboutDescription", "Chat command: |c66FF99"..HoW_RaMT.command.."|r\n\n|cEECA2A"..HoW_RaMT.displayName.."|r |c6699FF"..HoW_RaMT.version.."|r by |cFF9933"..HoW_RaMT.author.."|r")
end

function HoW_RaMT.IsVisible()
    return HoW_RaMT.vars and HoW_RaMT.vars.visible
end

function HoW_RaMT.SetVisible(checkbox)
    HoW_RaMT.vars.visible = checkbox
end

-- function HoW_RaMT.GetFont()
--     if HoW_RaMT.vars and HoW_RaMT.vars.font then
--         return HoW_RaMT.vars.font
--     else
--         return "ZoFontGame"
--     end
-- end

-- function HoW_RaMT.SetFont(font)
--     HoW_RaMT.vars.font = font
--     HoW_RaMT_CT_HeaderLabel:SetFont(font)
--     HoW_RaMT_CT_TextLabel:SetFont(font)
--     HoW_RaMT_CT_DirListTB:SetFont(font)
-- end

function HoW_RaMT.GetAlpha()
    if HoW_RaMT.vars and HoW_RaMT.vars.alpha then
        return HoW_RaMT.vars.alpha
    else
        return 65
    end
end

function HoW_RaMT.SetAlpha(alpha)
    -- store in vars
    HoW_RaMT.vars.alpha = alpha
    -- update ui
    HoW_RaMT_CT_Backdrop:SetAlpha(alpha / 100)
end

--------------------------------------------------------------------------------
-- Update functions.
--------------------------------------------------------------------------------

--- Update all timers for current character.
-- @return void
function HoW_RaMT.Update()
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_BLACKSMITHING]].timers = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_BLACKSMITHING)
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_CLOTHIER]].timers      = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_CLOTHIER)
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_WOODWORKING]].timers   = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_WOODWORKING)
    HoW_RaMT.vars.types[MOUNTS_TYPE[KEY_LANG]].timers                                 = HoW_RaMT.GetMountFeedTimers()
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
end

--- Update mount timers for current character.
-- @return void
function HoW_RaMT.UpdateMounts()
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
    HoW_RaMT.vars.types[MOUNTS_TYPE[KEY_LANG]].timers = HoW_RaMT.GetMountFeedTimers()
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
end

--- Update research timers for current character.
-- @return void
function HoW_RaMT.UpdateResearch()
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_BLACKSMITHING]].timers = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_BLACKSMITHING)
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_CLOTHIER]].timers      = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_CLOTHIER)
    HoW_RaMT.vars.types[CRAFTING_TYPES[KEY_LANG][CRAFTING_TYPE_WOODWORKING]].timers   = HoW_RaMT.GetCraftingTimers(CRAFTING_TYPE_WOODWORKING)
    -- d("-----[ "..HoW_RaMT.name.." ]----------------------------------------")
end

--- Get crafting timers for skill type.
-- @param craftingSkillType
-- @return a table with timers.
function HoW_RaMT.GetCraftingTimers(craftingSkillType)
    local maxSimultaneousResearch = GetMaxSimultaneousSmithingResearch(craftingSkillType)
    local numLines = GetNumSmithingResearchLines(craftingSkillType)
    local timestamp = GetTimeStamp()
    local timers = {}
    local index = 0
    for i = 1, numLines, 1 do
        local name, icon, numTraits, timeRequiredForNextResearchSecs = GetSmithingResearchLineInfo(craftingSkillType, i)
        for j = 1, numTraits, 1 do
            local duration, timeRemainingSecs = GetSmithingResearchLineTraitTimes(craftingSkillType, i, j)
            if duration and timeRemainingSecs then
                index = index + 1
                local traitType, traitDescription, known = GetSmithingResearchLineTraitInfo(craftingSkillType, i, j)
                table.insert(timers, {
                    ["index"] = index,
                    ["maxSlots"] = maxSimultaneousResearch,
                    ["name"] = name,
                    ["type"] = CRAFTING_TYPES[HoW_RaMT.gvars.language][craftingSkillType],
                    ["duration"] = duration,
                    ["remaining"] = timeRemainingSecs,
                    ["timestamp"] = timestamp,

                    ["traitType"] = ITEM_TRAIT_TYPES[HoW_RaMT.gvars.language][traitType],
                    ["traitDescription"] = traitDescription
                })
                -- d(string.format("[%s %d/%d] %s %s [%s]", CRAFTING_TYPES[HoW_RaMT.gvars.language][craftingSkillType], index, maxSimultaneousResearch, ITEM_TRAIT_TYPES[HoW_RaMT.gvars.language][traitType], name, FormatTimeSeconds(timeRemainingSecs, TIME_FORMAT_STYLE_DESCRIPTIVE_SHORT, TIME_FORMAT_PRECISION_SECONDS)))
            end
        end
    end
    if index == 0 then
        -- d(string.format("[%s %d/%d] No research in progress!", CRAFTING_TYPES[HoW_RaMT.gvars.language][craftingSkillType], index, maxSimultaneousResearch))
    end
    return timers
end

--- Get time to feed timers for all mounts.
-- @return a table with timers.
function HoW_RaMT.GetMountFeedTimers()
    local maxStableSlots = GetMaxStableSlots()
    local timestamp = GetTimeStamp()
    local timers = {}
    local index = 0
    for i = 1, maxStableSlots, 1 do
        if not IsStableSlotEmpty(i) then
            index = index + 1
            local name, iconTexture, portraitTexture, breed, description, sellPrice = GetStableSlotInfo(i)
            local level, inventoryBonus, maxInventoryBonus, staminaBonus, maxStaminaBonus, speedBonus, maxSpeedBonus = GetStableSlotMountStats(i)
            local timeMs, totalDurationMs = GetTimeUntilMountCanBeFed(i)
            if timeMs and totalDurationMs then
                table.insert(timers, {
                    ["index"] = index,
                    ["maxSlots"] = maxStableSlots,
                    ["name"] = name,
                    ["type"] = MOUNTS_TYPE[HoW_RaMT.gvars.language],
                    ["duration"] = totalDurationMs / 1000,
                    ["remaining"] = timeMs / 1000,
                    ["timestamp"] = timestamp,

                    ["breed"] = breed,
                    ["description"] = description,
                    ["level"] = level,
                    ["speedBonus"] = speedBonus,
                    ["staminaBonus"] = staminaBonus,
                    ["inventoryBonus"] = inventoryBonus
                })
                -- d(string.format("[Mount %d/%d] %s L%d %s [%s] [Speed: %d%%, Stamina: %d, Capacity: %d]", index, maxStableSlots, name, level, breed, FormatTimeMilliseconds(timeMs, TIME_FORMAT_STYLE_DESCRIPTIVE_SHORT, TIME_FORMAT_PRECISION_SECONDS), speedBonus, staminaBonus, inventoryBonus))
            else
                table.insert(timers, {
                    ["index"] = index,
                    ["maxSlots"] = maxStableSlots,
                    ["name"] = name,
                    ["type"] = MOUNTS_TYPE[HoW_RaMT.gvars.language],
                    ["duration"] = 0,
                    ["remaining"] = 0,
                    ["timestamp"] = timestamp,

                    ["breed"] = breed,
                    ["description"] = description,
                    ["level"] = level,
                    ["speedBonus"] = speedBonus,
                    ["staminaBonus"] = staminaBonus,
                    ["inventoryBonus"] = inventoryBonus
                })
                -- d(string.format("[Mount %d/%d] %s L%d %s is hungry! [Speed: %d%%, Stamina: %d, Capacity: %d]", index, maxStableSlots, name, level, breed, speedBonus, staminaBonus, inventoryBonus))
            end
        end
    end
    return timers
end

--------------------------------------------------------------------------------
-- Show timers window.
--------------------------------------------------------------------------------

--- Show timers window with all timers or timers for a type and/or character.
-- @param options table with options to show, types/character.
-- @return void
function HoW_RaMT.OpenWindow(options)
    -- Remember window state.
    HoW_RaMT.vars.visible = true

    -- Save in HoW_RaMT.types so UpdateWindow can access it.
    if (options and options.types) then
        HoW_RaMT.types = options.types
    end
    if (options and options.unitName and HoW_RaMT.gvars[options.unitName]) then
        HoW_RaMT.unitName = options.unitName
    end
    HoW_RaMT:UpdateWindow(true)
    HoW_RaMT_CT:SetHidden(false)
end

--- Update all timers or timers for a type and/or character.
--- Called from xml without args and OpenWindow.
-- @param force trigger update without buffering when opening up window to show content directly.
-- @return void
function HoW_RaMT:UpdateWindow(force)
    if not force and not BufferReached("HoW_RaMT.OnUpdate", 1) then return end

    local currentTimestamp = GetTimeStamp()
    local vars = HoW_RaMT.gvars.chars[HoW_RaMT.unitName]

    if vars.unit.level == 50 then
        HoW_RaMT_CT_TextLabel:SetText(string.format(
            "|c%s"..HoW_RaMT.abbrName.."\n|c%s%s\n|c%sVR|c%s%d |c%s%s %s\n",
            HoW_RaMT.gvars.colors.header,
            HoW_RaMT.gvars.colors.section,
            vars.unit.name,
            HoW_RaMT.gvars.colors.text,
            HoW_RaMT.gvars.colors.value,
            vars.unit.rank,
            HoW_RaMT.gvars.colors.text,
            vars.unit.race,
            vars.unit.class
        ))
    else
        HoW_RaMT_CT_TextLabel:SetText(string.format(
            "|c%s"..HoW_RaMT.abbrName.."\n|c%s%s\n|c%sLvl |c%s%d |c%s%s %s\n",
            HoW_RaMT.gvars.colors.header,
            HoW_RaMT.gvars.colors.section,
            vars.unit.name,
            HoW_RaMT.gvars.colors.text,
            HoW_RaMT.gvars.colors.value,
            vars.unit.level,
            HoW_RaMT.gvars.colors.text,
            vars.unit.race,
            vars.unit.class
        ))
    end

    for _, type in pairs(HoW_RaMT.types) do
        if vars.types[type].visible then
            local timers = vars.types[type].timers
            howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                "|c%s%s (|c%s%d|c%s/|c%s%d|c%s)",
                HoW_RaMT.gvars.colors.section,
                type,
                HoW_RaMT.gvars.colors.value,
                #timers,
                HoW_RaMT.gvars.colors.section,
                HoW_RaMT.gvars.colors.value,
                vars.types[type].maxSlots,
                HoW_RaMT.gvars.colors.section
            ))
            if timers and #timers > 0 then
                for index, timer in pairs(timers) do
                    local remaining = timer.remaining - (currentTimestamp - timer.timestamp)
                    if timer.type == "Mounts" then
                        howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                            "|c%s%s |c%s- lvl |c%s%d |c%s%s\nSpeed: |c%s%d%%|c%s, stamina: |c%s%d|c%s, capacity: |c%s%d",
                            HoW_RaMT.gvars.colors.value,
                            timer.name,
                            HoW_RaMT.gvars.colors.text,
                            HoW_RaMT.gvars.colors.value,
                            timer.level,
                            HoW_RaMT.gvars.colors.text,
                            timer.breed,
                            HoW_RaMT.gvars.colors.value,
                            timer.speedBonus,
                            HoW_RaMT.gvars.colors.text,
                            HoW_RaMT.gvars.colors.value,
                            timer.staminaBonus,
                            HoW_RaMT.gvars.colors.text,
                            HoW_RaMT.gvars.colors.value,
                            timer.inventoryBonus
                        ))
                        if remaining > 0 then
                            howui:AppendText(HoW_RaMT_CT_TextLabel, HoW_RaMT:SecondsToAge(remaining))
                        elseif (timer.level < 50) then
                            howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                                "|c%sPlease feed me.",
                                HoW_RaMT.gvars.colors.alert
                            ))
                        end
                    else
                        howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                            "|c%s%s %s",
                            HoW_RaMT.gvars.colors.text,
                            timer.traitType,
                            timer.name
                        ))
                        if remaining > 0 then
                            howui:AppendText(HoW_RaMT_CT_TextLabel, HoW_RaMT:SecondsToAge(remaining))
                        else
                            howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                                "|c%sResearch completed.",
                                HoW_RaMT.gvars.colors.info
                            ))
                        end
                    end
                end
            else
                howui:AppendText(HoW_RaMT_CT_TextLabel, string.format(
                    "|c%sNo research in progress.",
                    HoW_RaMT.gvars.colors.alert
                ))
            end
            howui:AppendText(HoW_RaMT_CT_TextLabel, "")
        end
    end
end

--- Convert seconds to an age string.
--- Similar to FormatTimeSeconds(remaining, TIME_FORMAT_STYLE_DESCRIPTIVE_SHORT, TIME_FORMAT_PRECISION_SECONDS)
-- @param seconds value to convert
-- @return string with age
function HoW_RaMT:SecondsToAge(seconds)
    local d = math.floor(seconds / 86400)
    local h = math.floor((seconds - (d * 86400)) / 3600)
    local m = math.floor((seconds - (d * 86400) - (h * 3600)) / 60)
    local s = seconds - (d * 86400) - (h * 3600) - (m * 60)
    if (d > 0) then
        if (d > 1) then
            return string.format(
                "|c%s%d |c%sdays |c%s%d |c%shrs |c%s%d |c%smin |c%s%d |c%ssec",
                HoW_RaMT.gvars.colors.value,
                d,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                h,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                m,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                s,
                HoW_RaMT.gvars.colors.text
            )
        else
            return string.format(
                "|c%s%d |c%sday |c%s%d |c%shrs |c%s%d |c%smin |c%s%d |c%ssec",
                HoW_RaMT.gvars.colors.value,
                d,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                h,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                m,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                s,
                HoW_RaMT.gvars.colors.text
            )
        end
    end
    if (h > 0) then
        if (h > 1) then
            return string.format(
                "|c%s%d |c%shrs |c%s%d |c%smin |c%s%d |c%ssec",
                HoW_RaMT.gvars.colors.value,
                h,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                m,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                s,
                HoW_RaMT.gvars.colors.text
            )
        else
            return string.format(
                "|c%s%d |c%shr |c%s%d |c%smin |c%s%d |c%ssec",
                HoW_RaMT.gvars.colors.value,
                h,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                m,
                HoW_RaMT.gvars.colors.text,
                HoW_RaMT.gvars.colors.value,
                s,
                HoW_RaMT.gvars.colors.text
            )
        end
    end
    if (m > 0) then
        return string.format(
            "|c%s%d |c%smin |c%s%d |c%ssec",
            HoW_RaMT.gvars.colors.value,
            m,
            HoW_RaMT.gvars.colors.text,
            HoW_RaMT.gvars.colors.value,
            s,
            HoW_RaMT.gvars.colors.text
        )
    end
    return string.format(
        "|c%s%d |c%ssec",
        HoW_RaMT.gvars.colors.value,
        s,
        HoW_RaMT.gvars.colors.text
    )
end

--------------------------------------------------------------------------------
-- Update xml control functions.
--------------------------------------------------------------------------------

--- Set window anchor position.
-- @param anchor table with point, relativePoint, offsetX and offsetY.
-- @return void
function HoW_RaMT:SetAnchor(anchor)
    HoW_RaMT_CT:SetAnchor(anchor.point, GuiRoot, anchor.relativePoint, anchor.offsetX, anchor.offsetY)
end

--- Set window size.
-- @param dimension table with width and height
-- @return void
function HoW_RaMT:SetDimensions(dimension)
    HoW_RaMT_CT:SetDimensions(dimension.width, dimension.height)
end

--- Set window faded.
-- @param faded boolean faded state
-- @return void
function HoW_RaMT:SetFaded(faded)
    HoW_RaMT_CT_Backdrop:SetHidden(faded)
end


--- Fade window backdrop.
-- @return void
function HoW_RaMT:FadeWindow()
    -- Remember window faded state.
    HoW_RaMT.vars.faded = not HoW_RaMT_CT_Backdrop:IsHidden()

    HoW_RaMT:SetFaded(HoW_RaMT.vars.faded)
end

--- Close the window.
-- @return void
function HoW_RaMT:CloseWindow()
    -- Remember window state.
    HoW_RaMT.vars.visible = false

    HoW_RaMT_CT:SetHidden(true)
end

--- Hide the window if visible and screen state changed to some dialogue/menu interface.
-- @return void
function HoW_RaMT:HideIfVisible()
    if HoW_RaMT.vars.visible then
        HoW_RaMT_CT:SetHidden(true)
    end
end

--- Show the window if visible and screen state returned to normal.
-- @return void
function HoW_RaMT:ShowIfVisible()
    if HoW_RaMT.vars.visible then
        HoW_RaMT_CT:SetHidden(false)
    end
end

--- Add buttons for each character.
-- @return void
function HoW_RaMT:AddCharacterButtons()
    local i = 1

    -- Reference point for first button.
    local button = HoW_RaMT_CT_CloseButton
    local anchor = {point = TOPRIGHT, relativeTo = button, relativePoint = BOTTOMRIGHT, offsetX = 0, offsetY = 30}
    local dimension = {width = 24, height = 24}
    for _, v in pairs(HoW_RaMT.gvars.names) do
        local unit = HoW_RaMT.gvars.chars[v].unit
        button = howui:Button(
            HoW_RaMT.prefix.."CharButton"..i,
            "",
            dimension,
            HoW_RaMT_CT,
            anchor
        )
        howui:Texture(
            button:GetName() .. "_Texture",
            "/esoui/art/contacts/social_classicon_"..unit.class..".dds",
            button
        )
        HoW_RaMT.SetToolTip(button, string.format("|cFFE3A1%s\n|cEEEEEELevel %d %s %s", unit.name, unit.level, unit.race, unit.class))
        button:SetHandler("OnClicked", function(self)
            HoW_RaMT.unitName = unit.name
            local vars = HoW_RaMT.gvars.chars[HoW_RaMT.unitName]
            -- HoW_RaMT:SetAnchor(vars.anchor)
            HoW_RaMT:SetDimensions(vars.dimension)
            HoW_RaMT:UpdateWindow(true)
        end)
        -- Prepare for next iteration
        i = i + 1
        anchor = {point = TOPRIGHT, relativeTo = button, relativePoint = BOTTOMRIGHT, offsetX = 0, offsetY = 0}
    end
end

--- Add a tooltip to a control.
-- @param ctrl the control element.
-- @param text the text to show on mouse over.
-- @return void
function HoW_RaMT.SetToolTip(ctrl, text)
    ctrl:SetHandler("OnMouseEnter", function(self)
        ZO_Tooltips_ShowTextTooltip(self, TOP, text)
    end)
    ctrl:SetHandler("OnMouseExit", function(self)
        ZO_Tooltips_HideTextTooltip()
    end)
end

--- Show/hide mounts info.
-- @return void
function HoW_RaMT:ToggleMounts()
    local type = HoW_RaMT.gvars.chars[HoW_RaMT.unitName].types.Mounts
    if type.visible then
        HoW_RaMT_CT_MountsButton_Texture:SetTexture("/esoui/art/mounts/tabicon_mounts_up.dds")
        type.visible = false
    else
        HoW_RaMT_CT_MountsButton_Texture:SetTexture("/esoui/art/mounts/tabicon_mounts_down.dds")
        type.visible = true
    end
    HoW_RaMT:UpdateWindow(true)
end

--- Show/hide blacksmithing info.
-- @return void
function HoW_RaMT:ToggleBlacksmithing()
    local type = HoW_RaMT.gvars.chars[HoW_RaMT.unitName].types.Blacksmithing
    if type.visible then
        type.visible = false
    else
        type.visible = true
    end
    HoW_RaMT:UpdateWindow(true)
end

--- Show/hide clothing info.
-- @return void
function HoW_RaMT:ToggleClothing()
    local type = HoW_RaMT.gvars.chars[HoW_RaMT.unitName].types.Clothing
    if type.visible then
        type.visible = false
    else
        type.visible = true
    end
    HoW_RaMT:UpdateWindow(true)
end

--- Show/hide blacksmithing info.
-- @return void
function HoW_RaMT:ToggleWoodworking()
    local type = HoW_RaMT.gvars.chars[HoW_RaMT.unitName].types.Woodworking
    if type.visible then
        type.visible = false
    else
        type.visible = true
    end
    HoW_RaMT:UpdateWindow(true)
end

--------------------------------------------------------------------------------

-- Initialize addon.
EVENT_MANAGER:RegisterForEvent(HoW_RaMT.name, EVENT_ADD_ON_LOADED, HoW_RaMT.AddOnLoaded)
